<?php

	/**
	 * Erlaubt den Verkauf von Abos
	 * @author Daschmi
	 */
	class wpsg_mod_abo extends wpsg_mod_basic {

		var $lizenz = 1;
		var $id = 1630;
		var $hilfeURL = 'http://wpshopgermany.de/?p=3292';
		var $version = "9.9.9";

		const CONFIG = '-1';

		const EXPIRE_NONE = '0';
        const EXPIRE_MAIL = '1';
        const EXPIRE_AUTO = '2';

        const CUSTOMER_AUTO_NOTHING = '0'; // Kunde kann im Frontend nicht auf die automatische Verlängeurng eingreifen
        const CUSTOMER_AUTO_DIRECT = '1'; // Kunde kann die automatische Verlängerung direkt stoppen
        const CUSTOMER_AUTO_MAIL = '2'; // Kunde kann die automatische Verlängerung über eine Mail an den Admin stoppen

        const STATUS_ACTIVE = '0'; // Standard
        const STATUS_STOPPED = '1'; // Abo ist gestoppt, es erfolgt keine manuelle oder automatische Verlängerung

		/**
		 * Costructor
		 */
		public function __construct() {

			parent::__construct();

			$this->name = __('AboProdukte', 'wpsg');
			$this->group = __('Produkte', 'wpsg');
			$this->desc = __('Ermöglicht den Verkauf von Abo Produkten', 'wpsg');

		} // public function __construct()

        public function init() {

            // Option gab es früher nicht und wird hier auf den Default Wert korrigiert.
            if ($this->shop->get_option('wpsg_mod_abo_expiremode') === false) $this->shop->update_option('wpsg_mod_abo_expiremode', self::EXPIRE_MAIL);

            require_once WPSG_PATH_MOD.'/mod_abo/wpsg_abo.php';

			$role_object = get_role('administrator');
			$role_object->add_cap('wpsg_abo');

        } // public function init()

        public function admin_setcapabilities() {

        	$this->render(WPSG_PATH_VIEW.'/mods/mod_abo/admin_setcapabilities.phtml');

        } // public function admin_setcapabilities()

		public function install() {

			require_once(WPSG_PATH_WP.'/wp-admin/includes/upgrade.php');

			// Grundeinstellung Pro Seite
			if ($this->shop->get_option('wpsg_mod_abo_perpage') === false || $this->shop->get_option('wpsg_mod_abo_perpage') == '') {

				$this->shop->update_option('wpsg_mod_abo_perpage', '20');

			}

			/**
			 * Produkttabelle erweitern
			*/
			$sql = "CREATE TABLE ".WPSG_TBL_PRODUCTS." (
			  	wpsg_mod_abo_activ int(1) NOT NULL,
				wpsg_mod_abo_resell int(11) NOT NULL,
				wpsg_mod_abo_price DOUBLE(10,2) NOT NULL,
				wpsg_mod_abo_durration int(11) NOT NULL,
				wpsg_mod_abo_url varchar(500) NOT NULL
			) ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;";

			dbDelta($sql);

			/**
			 * Abo Tabelle anlegen
			 */
			$sql = "CREATE TABLE ".WPSG_TBL_ABO." (
				id mediumint(9) NOT NULL AUTO_INCREMENT,
	  			cdate datetime NOT NULL,
				order_id int(11) NOT NULL,
				order_product_id int(11) NOT NULL, 
				expiration DATETIME NOT NULL,
				inform_expiration int(1) DEFAULT '0' NOT NULL,
				inform_autorenew int(1) DEFAULT '0' NOT NULL COMMENT 'Auf 1 wenn die Mail X-Tage vor verlängerung gesendet wurde',
				rebuy int(11) DEFAULT '0' NOT NULL,
				status int(1) DEFAULT '0' NOT NULL,
				mail_admin_stop DATETIME NOT NULL COMMENT 'Wird auf 1 gesetzt, wenn der Admin das Abo stoppen soll',
				product_key VARCHAR(255) NOT NULL COMMENT 'Deprecated, wird aber wegen Abwärtskompatibilität noch gebraucht',
				PRIMARY KEY  (id),
	  			KEY order_id (order_id), 
				KEY order_product_id (order_product_id)
			) ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;";

			dbDelta($sql);

			/**
			 * Bestelltabelle erweitern um bei Folgebestellungen zu wissen welche die Originalbestellung ist
			 */
			$sql = "CREATE TABLE ".WPSG_TBL_ORDER." (
				wpsg_mod_abo_origin_order int(11) NOT NULL,
				KEY wpsg_mod_abo_origin_order (wpsg_mod_abo_origin_order)
			) ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;";

			dbDelta($sql);

			$this->shop->checkDefault('wpsg_wpsgmodabofirstbuy_betreff', __('Sie haben erfolgreich ein Abo erworben', 'wpsg'));
			$this->shop->checkDefault('wpsg_wpsgmodaboexpiration_betreff', __('Ihr Abo ist ausgelaufen', 'wpsg'));
			$this->shop->checkDefault('wpsg_wpsgmodabrebuy_betreff', __('Ihr Abo ist verlängert worden', 'wpsg'));

			$this->shop->checkDefault('wpsg_mod_abo_wpsgmodabofirstbuy', '1');
			$this->shop->checkDefault('wpsg_mod_abo_wpsgmodabrebuy', '1');

		} // public function install()

        public function delOrder($order_id) {

            // Alle Abo-Produkte zu dieser Bestellung in der Tabelle WP_WPSG_ABO löschen
            $this->db->Query("DELETE FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."' ");

        } // public function delOrder($order_id)

		/**
		 * Übersicht über die Kunden im Backend
		 */
		private function be_indexAction() {

			if (isset($_REQUEST['submit-button'])) wpsg_checkNounce('Abo', 'index', ['search' => '1']);

			$nPerPage = $this->shop->get_option('wpsg_mod_abo_perpage');
			if ($nPerPage <= 0) $nPerPage = 10;

			$this->shop->view['hasFilter'] = false;
			$this->shop->view['pages'] = 0;
			$this->shop->view['arState'] = [];
			$this->shop->view['arFilter'] = array(
                'order' => 'cdate',
                'ascdesc' => 'DESC',
                'page' => '1'
			);

			if (isset($_REQUEST['filter']['s'])) $this->shop->view['arFilter']['s'] = sanitize_text_field($_REQUEST['filter']['s']);
            if (isset($_REQUEST['filter']['k_id'])) $this->shop->view['arFilter']['k_id'] = sanitize_text_field($_REQUEST['filter']['k_id']);
            if (isset($_REQUEST['filter']['cdate_m'])) $this->shop->view['arFilter']['cdate_m'] = intval($_REQUEST['filter']['cdate_m']);
            if (isset($_REQUEST['filter']['cdate_y'])) $this->shop->view['arFilter']['cdate_y'] = intval($_REQUEST['filter']['cdate_y']);
            if (isset($_REQUEST['filter']['enddate_m'])) $this->shop->view['arFilter']['enddate_m'] = intval($_REQUEST['filter']['enddate_m']);
            if (isset($_REQUEST['filter']['enddate_y'])) $this->shop->view['arFilter']['enddate_y'] = intval($_REQUEST['filter']['enddate_y']);
            if (isset($_REQUEST['filter']['order'])) $this->shop->view['arFilter']['order'] = $_REQUEST['filter']['order'];
            if (isset($_REQUEST['filter']['ascdesc'])) $this->shop->view['arFilter']['ascdesc'] = $_REQUEST['filter']['ascdesc'];
            if (wpsg_isSizedInt($_REQUEST['filter']['status'])) $this->shop->view['arFilter']['state'] = intval($_REQUEST['filter']['status']);

            $filter_state = $this->shop->view['arFilter'];
            unset($filter_state['state']);

            $this->shop->view['arState'][0] = [
                'label' => __('Alle', 'wpsg'),
                'count' => wpsg_abo::count($filter_state)
            ];
 
			foreach (wpsg_abo::$arState as $state_key => $state_label) {

			    $filter_state = $this->shop->view['arFilter'];
			    $filter_state['state'] = intval($state_key);
				
			    $count = wpsg_abo::count($filter_state);
				
			    if ($count > 0) {

			        $this->shop->view['arState'][$state_key] = [
                        'label' => $state_label,
                        'count' => wpsg_abo::count($filter_state)
                    ];

                }

            } 

			$this->shop->view['countAll'] = wpsg_abo::count($this->shop->view['arFilter']);

			if (wpsg_isSizedInt($_REQUEST['seite'])) $this->shop->view['arFilter']['page'] = $_REQUEST['seite'];

			$this->shop->view['pages'] = ceil($this->shop->view['countAll'] / $nPerPage);
			if (!isset($this->shop->view['arFilter']['page']) || $this->shop->view['arFilter']['page'] <= 0 || $this->shop->view['arFilter']['page'] > $this->shop->view['pages']) $this->shop->view['arFilter']['page'] = 1;

			$this->shop->view['arFilter']['limit'] = array(($this->shop->view['arFilter']['page'] - 1) * $nPerPage, $nPerPage);

			// Filter speichern
			$_SESSION['wpsg']['backend']['order']['arFilter'] = $this->shop->view['arFilter'];

			$this->shop->view['arData'] = wpsg_abo::find($this->shop->view['arFilter']);
			
			$this->shop->view['cdate_years'] = $this->db->fetchAssocField("SELECT DISTINCT DATE_FORMAT(`cdate`, '%Y') FROM `".WPSG_TBL_ORDER."` ORDER BY `cdate` ASC ");

			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/index.phtml');

		} // private function be_indexAction()

		public function dispatch() {

		    if (wpsg_isSizedString($_REQUEST['action'], 'delete')) {

		        $abo_id = intval($_REQUEST['abo_id']);

                $this->db->Query("DELETE FROM `".WPSG_TBL_ABO."` WHERE `id` = '".wpsg_q($abo_id)."' ");

                $this->shop->addBackendMessage(__('Abo erfolgreich gelöscht.', 'wpsg'));
                $this->shop->redirect(WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Abo');

            }

			$this->be_indexAction();

		} // public function dispatch()

		public function wpsg_add_pages($default_page)
		{

			add_submenu_page($default_page, __("Aboverwaltung", "wpsg"), __("Aboverwaltung", "wpsg"), 'wpsg_abo', 'wpsg-Abo', array($this, 'dispatch'));

		} // public function wpsg_add_pages()

		public function settings_edit()
		{

			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/settings_edit.phtml');

		} // public function settings_edit()

		public function settings_save() {

			$this->shop->update_option('wpsg_mod_abo_url', $_REQUEST['wpsg_mod_abo_url'], false, false, WPSG_SANITIZE_URL);
			$this->shop->update_option('wpsg_mod_abo_perpage', $_REQUEST['wpsg_mod_abo_perpage'], false, false, WPSG_SANITIZE_INT);

			$this->shop->update_option('wpsg_mod_abo_wpsgmodabofirstbuy', $_REQUEST['wpsg_mod_abo_wpsgmodabofirstbuy'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_abo_wpsgmodabrebuy', $_REQUEST['wpsg_mod_abo_wpsgmodabrebuy'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_abo_expiremode', $_REQUEST['wpsg_mod_abo_expiremode'], false, false, WPSG_SANITIZE_VALUES, [wpsg_mod_abo::EXPIRE_NONE, wpsg_mod_abo::EXPIRE_AUTO, wpsg_mod_abo::EXPIRE_MAIL]);
			$this->shop->update_option('wpsg_mod_abo_premode', $_REQUEST['wpsg_mod_abo_premode'], false, false, WPSG_SANITIZE_VALUES, [wpsg_mod_abo::EXPIRE_NONE, wpsg_mod_abo::EXPIRE_AUTO, wpsg_mod_abo::EXPIRE_MAIL]);

			$this->shop->update_option('wpsg_mod_abo_autoexpire_payment', $_REQUEST['wpsg_mod_abo_autoexpire_payment'], false, false, WPSG_SANITIZE_VALUES, array_merge(['origin'], array_keys($this->shop->arPayment)));
			$this->shop->update_option('wpsg_mod_abo_autoexpire_customermail', $_REQUEST['wpsg_mod_abo_autoexpire_customermail'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_abo_autoexpire_adminmail', $_REQUEST['wpsg_mod_abo_autoexpire_adminmail'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_abo_autoexpire_customer', $_REQUEST['wpsg_mod_abo_autoexpire_customer'], false, false, WPSG_SANITIZE_VALUES, [wpsg_mod_abo::CUSTOMER_AUTO_DIRECT, wpsg_mod_abo::CUSTOMER_AUTO_MAIL, wpsg_mod_abo::CUSTOMER_AUTO_NOTHING]);
			$this->shop->update_option('wpsg_mod_abo_autoexpire_informationtime', wpsg_xss($_REQUEST['wpsg_mod_abo_autoexpire_informationtime']), false, false, WPSG_SANITIZE_TEXTFIELD);

		} // public function settings_save()

		public function product_getPrice(&$oProduct, &$price_netto, &$price_brutto, $product_key, $amount, $weight) {

			if (preg_match('/^abo_\d+\_\d+$/', $product_key)) {

				$arKey = explode('_', $product_key);
				$oOrderProduct = wpsg_order_product::getInstance(intval($arKey[2]));

				if ($this->shop->getBackendTaxview() === WPSG_NETTO) {

					$price_netto = $oOrderProduct->getProduct()->__get('wpsg_mod_abo_price');
					$price_brutto = wpsg_calculatePreis($price_netto, WPSG_BRUTTO, $this->shop->getDefaultCountry()->getTax($oOrderProduct->getProduct()->mwst_key));

				} else {

					$price_brutto = $oOrderProduct->getProduct()->__get('wpsg_mod_abo_price');
					$price_netto = wpsg_calculatePreis($price_brutto, WPSG_NETTO, $this->shop->getDefaultCountry()->getTax($oOrderProduct->getProduct()->mwst_key));

				}

			}

		}

		public function product_addedit_content(&$product_content, &$product_data) {

			if (!wpsg_isSizedInt($product_data['id']) || isset($_REQUEST['wpsg_lang'])) return false;

			if ($product_data['wpsg_mod_abo_resell'] > 0)
			{

				$product_data['wpsg_mod_abo_resell_count'] = $product_data['wpsg_mod_abo_resell'];
				$product_data['wpsg_mod_abo_resell'] = 1;

			}
			else
			{

				$product_data['wpsg_mod_abo_resell_count'] = '1';
				$product_data['wpsg_mod_abo_resell'] = 0;

			}

			$this->shop->view['product'] = $product_data;

		    $oProduct = wpsg_product::getInstance($product_data['id']);

            $this->shop->view['product']['wpsg_mod_abo_expiremode'] = $oProduct->getMeta('wpsg_mod_abo_expiremode', false, wpsg_mod_abo::CONFIG);
            $this->shop->view['product']['wpsg_mod_abo_premode'] = $oProduct->getMeta('wpsg_mod_abo_premode', false, wpsg_mod_abo::CONFIG);
            $this->shop->view['product']['wpsg_mod_abo_autoexpire_payment'] = $oProduct->getMeta('wpsg_mod_abo_autoexpire_payment', false, wpsg_mod_abo::CONFIG);
            $this->shop->view['product']['wpsg_mod_abo_autoexpire_customermail'] = $oProduct->getMeta('wpsg_mod_abo_autoexpire_customermail', false, wpsg_mod_abo::CONFIG);
            $this->shop->view['product']['wpsg_mod_abo_autoexpire_adminmail'] = $oProduct->getMeta('wpsg_mod_abo_autoexpire_adminmail', false, wpsg_mod_abo::CONFIG);
            $this->shop->view['product']['wpsg_mod_abo_autoexpire_informationtime'] = $oProduct->getMeta('wpsg_mod_abo_autoexpire_informationtime', false, '');
            $this->shop->view['product']['wpsg_mod_abo_autoexpire_customer'] = $oProduct->getMeta('wpsg_mod_abo_autoexpire_customer', false, wpsg_mod_abo::CONFIG);

			$product_content['wpsg_mod_abo'] = array(
					'title' => __('Abo', 'wpsg'),
					'content' => $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/produkt_addedit_content.phtml', false)
			);

		} // public function product_addedit_content(&$product_content, &$produkt_data)

		public function produkt_save_before(&$product_data) {

			if (isset($_REQUEST['wpsg_lang'])) return false;

			if (isset($_REQUEST['wpsg_mod_abo_activ']) && $_REQUEST['wpsg_mod_abo_activ'] == '1') {

				$product_data['wpsg_mod_abo_activ'] = '1';

				wpsg_checkRequest('wpsg_mod_abo_price', [WPSG_SANITIZE_FLOAT], __('Abo Produkt / Abo Preis', 'wpsg'), $product_data);
				wpsg_checkRequest('wpsg_mod_abo_durration', [WPSG_SANITIZE_INT], __('Abo Produkt / Abo Dauer', 'wpsg'), $product_data);
				wpsg_checkRequest('wpsg_mod_abo_url', [WPSG_SANITIZE_URL], __('Abo Produkt / URL Benachrichtigung', 'wpsg'), $product_data);

				if ($_REQUEST['wpsg_mod_abo_resell'] == '1') {

					if ($_REQUEST['wpsg_mod_abo_resell_count'] <= 0) {

						$this->shop->addBackendError(__('Ungültige Anzahl von zulässigen Abokäufen. Wurde auf unendlich gesetzt.', 'wpsg'));

						$product_data['wpsg_mod_abo_resell'] = '0';

					} else {

						$product_data['wpsg_mod_abo_resell'] = wpsg_q(wpsg_sinput("key", $_REQUEST['wpsg_mod_abo_resell_count']));

					}

				} else {

					$product_data['wpsg_mod_abo_resell'] = '0';

				}

			} else {

				$product_data['wpsg_mod_abo_activ'] = '0';

			}

		} // public function produkt_save(&$produkt_id)

        public function produkt_save(&$product_id) {

		    $oProduct = wpsg_product::getInstance($product_id);

		    if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_expiremode'], WPSG_SANITIZE_VALUES, [self::CONFIG, self::EXPIRE_AUTO, self::EXPIRE_MAIL]))
		        $oProduct->setMeta('wpsg_mod_abo_expiremode', $_REQUEST['wpsg_mod_abo_expiremode']);

		    if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_premode'], WPSG_SANITIZE_VALUES, [self::CONFIG, self::EXPIRE_AUTO, self::EXPIRE_MAIL]))
		    	$oProduct->setMeta('wpsg_mod_abo_premode', $_REQUEST['wpsg_mod_abo_premode']);

		    if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_autoexpire_payment'], WPSG_SANITIZE_VALUES, array_merge(['origin'], array_keys($this->shop->arPayment)))) $oProduct->setMeta('wpsg_mod_abo_autoexpire_payment', $_REQUEST['wpsg_mod_abo_autoexpire_payment']);
            if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_autoexpire_customermail'], WPSG_SANITIZE_VALUES, [self::CONFIG, '0', '1'])) $oProduct->setMeta('wpsg_mod_abo_autoexpire_customermail', $_REQUEST['wpsg_mod_abo_autoexpire_customermail']);
            if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_autoexpire_adminmail'], WPSG_SANITIZE_VALUES, [self::CONFIG, '0', '1'])) $oProduct->setMeta('wpsg_mod_abo_autoexpire_adminmail', $_REQUEST['wpsg_mod_abo_autoexpire_adminmail']);
			if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_autoexpire_customer'], WPSG_SANITIZE_VALUES, [wpsg_mod_abo::CONFIG, wpsg_mod_abo::CUSTOMER_AUTO_NOTHING, wpsg_mod_abo::CUSTOMER_AUTO_DIRECT, wpsg_mod_abo::CUSTOMER_AUTO_MAIL])) $oProduct->setMeta('wpsg_mod_abo_autoexpire_customer', $_REQUEST['wpsg_mod_abo_autoexpire_customer']);
		    if (wpsg_checkInput($_REQUEST['wpsg_mod_abo_autoexpire_informationtime'], WPSG_SANITIZE_INT, ['allowEmpty' => true])) $oProduct->setMeta('wpsg_mod_abo_autoexpire_informationtime', $_REQUEST['wpsg_mod_abo_autoexpire_informationtime']);

        }

		public function setAboStatus($order_id, $status_id, $inform)
		{

			$this->db->UpdateQuery(WPSG_TBL_ABO, array(
					'status' => $status_id
			), "`order_id` = '".wpsg_q($order_id)."'");

			$oOrder = wpsg_order::getInstance($order_id);
			$this->shop->view['stornoDate'] = time();

			$this->shop->view['customer']['vname'] = $oOrder->getInvoiceFirstName();;
			$this->shop->view['customer']['name'] = $oOrder->getInvoiceName();;
			$this->shop->view['customer']['email'] = $oOrder->getCustomer()->getEMail();

			$arOrderProducts = $this->getAboProductsFromOrder($order_id);

			foreach ($arOrderProducts as $oOrderProduct)
			{

				$abo_status = $this->getAboProductStatus($order_id, $oOrderProduct->getProductId());

				if ($inform) $this->sendCustomerMailResign($abo_status);

			}

		}	// public function setAboStatus($order_id, $status_id, $inform)

		public function setOrderStatus($order_id, $status_id, $inform) {

			if ($status_id == 100) {

				// Sind in der Bestellung Abo Produkte?
				$arBasket = $this->shop->cache->loadBasketArray($order_id);

				foreach ($arBasket['produkte'] as $product) {

					$product_data = $this->shop->cache->loadProduct($this->shop->getProduktID($product['productkey']));

					if (substr($product['productkey'], 0, 3) == 'abo') {

					    // Kauf einer Verlängerung
                        // Hier ist der Productkey abo_{original_order_id}_{original_order_product_id}

						// BestellID der OriginalBestellung
						$order_base_id = $this->getOrderIDFromKey($product['productkey']);
						$order_product_id = $this->getOrderProductIDFromKey($product['productkey']);

						// Eine Abo Verlängerung wurde gekauft
						$abo_entry = $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_base_id)."' AND `order_product_id` = '".wpsg_q($order_product_id)."'");

						if ($abo_entry['id'] <= 0) throw new \wpsg\Exception(__('Es wurde eine Bestellung auf Zahlung aktzeptiert gesetzt, die eine Abo Verlängerung drin hat zu der kein Eintrag existiert', 'wpsg'));
						if ($product_data['wpsg_mod_abo_resell'] != 0 && $abo_entry['rebuy'] + $product['menge'] > $product_data['wpsg_mod_abo_resell']) throw new \wpsg\Exception('Es wurde eine Bestellung auf Zahlung aktzeptiert gesetzt, das Abo konnte aber nicht verlängert werden da es die gültige Anzahl im Produkt überschreibt', __('wpsg'));

						if (strtotime($abo_entry['expiration']) < time()) {

							// Das Abo ist ausgelaufen
							$abo_expiration = $this->getAboExpiration($product_data['id'], $product['menge'], time(), true);

						} else {

							// Abo wurde verlängert und der Zeitraum wird an die alte Verlängerung angehängt
							$abo_expiration = $this->getAboExpiration($product_data['id'], $product['menge'], strtotime($abo_entry['expiration']), true);

						}

						$this->db->UpdateQuery(WPSG_TBL_ABO, array(
							'rebuy' => wpsg_q($abo_entry['rebuy'] + $product['menge']),
							'expiration' => $abo_expiration,
							'inform_expiration' => '0',
                            'inform_autorenew' => '0'
						), "`id` = '".wpsg_q($abo_entry['id'])."'");

						$url = $this->getNotifyURL($product_data['id']);

						if (wpsg_isSizedString($url)) {

							$this->shop->notifyURL($url, $product['productkey'], 1, $order_id, 3, array(
								'typ' => 'abo_rebuy',
								'abo_baseOrderID' => $order_base_id,
								'abo_kaufdatum' => date('Y-m-d H:i:s', time()),
								'abo_expiration' => $abo_expiration,
							));

						}

						if ($inform == true) {

							// Mail bei Verlängerung
							$this->sendAboRebuyMail($this->getOrderIDFromKey($product['productkey']), $order_product_id);

						}

					} else {

						if ($product_data['wpsg_mod_abo_activ'] == '1') {

							// Existiert ein Eintrag in der Abotabelle schon, so wird dieser gelöscht
							$abo_id = $this->db->fetchOne("SELECT `id` FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."' AND `order_product_id` = '".wpsg_q($product['order_product_id'])."'");
							if ($abo_id > 0) { $this->db->Query("DELETE FROM `".WPSG_TBL_ABO."` WHERE `id` = '".wpsg_q($abo_id)."'"); }

							// Es ist ein Abo Produkt in der Bestellung
							$abo_expiration = $this->getAboExpiration($product_data['id'], $product['menge'], time(), true);

							$abo_data_db = array(
								'cdate' => 'NOW()',
								'order_id' => wpsg_q($order_id),
								'order_product_id' => wpsg_q($product['order_product_id']),
								'rebuy' => '0',
								'expiration' => wpsg_q($abo_expiration),
								'inform_expiration' => '0'
							);

							$this->db->ImportQuery(WPSG_TBL_ABO, $abo_data_db);

							$url = $this->getNotifyURL($product_data['id']);

							if (wpsg_isSizedString($url)) {

								$this->shop->notifyURL($url, $product['productkey'], 1, $order_id, 3, array(
									'typ' => 'abo_firstbuy',
									'abo_kaufdatum' => date('Y-m-d H:i:s', time()),
									'abo_expiration' => $abo_expiration,
								));

							}

							if ($inform == true) {

								// Mail bei Erstkauf
								$this->sendAboFirstBuyMail($order_id, $product['order_product_id']);

							}

						}

					}

				}

			}

		}

		public function calculation_saveOrder(&$oCalculation, $arCalculation, &$db_data, $finish_order) {

			if (!$finish_order) return;

			foreach ($_SESSION['wpsg']['basket'] as $p) {

				if (substr($p['id'], 0, 3) == 'abo')
				{

					// Ein Abo Produkt wurde gekauft
					$arAbo = explode('_', $p['id']);

					$db_data['wpsg_mod_abo_origin_order'] = wpsg_q($arAbo[1]);

				}

			}

		}

		public function cron() {

			date_default_timezone_set('Europe/Berlin');
			
			$upload_dir = wp_upload_dir();
			$pid_file = $upload_dir['basedir'].'/wpsg_cron_abo';
			
			file_put_contents($upload_dir['basedir'].'/wpsg_cron_abo_log', date('d.m.Y H:i:s').' '.getmypid().' '.$_SERVER['REQUEST_URI'].' '.file_exists($pid_file)."\r\n", FILE_APPEND);
			
			if (file_exists($pid_file)) {
				
				echo "Cron läuft seit ".date('d.m.Y H:i:s', intval(file_get_contents($pid_file)))."\r\n";
				exit;
				
			} else {
				
				echo "Cron gestartet um ".date('d.m.Y H:i:s')."<br />";
				
				file_put_contents($pid_file, time());
				
			}
			
			ob_end_flush();
			
			$t_start = microtime(true);
			 
			$checkEndCron = function() use ($t_start, $pid_file) {
			
				if (($GLOBALS['wpsg_mod_abo_cron_count']??0) >= intval($_REQUEST['limit']??50)) {
					
					$this->shop->update_option('wpsg_mod_abo_cron_lastrun', time());
					$this->shop->update_option('wpsg_mod_abo_cron_run', '0');
		
					unlink($pid_file); 
					
					echo "Cron beendet weil Limit (".intval($_REQUEST['limit']??50).") erreicht<br />";
					exit;
					
				}
				
			};
						
			// Parallele Ausführung verhindern
			//$wpsg_mod_abo_cron_lastrun = intval($this->shop->get_option('wpsg_mod_abo_cron_lastrun'));			
			//if ($wpsg_mod_abo_cron_lastrun > 0 && $wpsg_mod_abo_cron_lastrun > time() - 120 * 60) return;
			if (intval($this->shop->get_option('wpsg_mod_abo_cron_run')) === 1) return;			
			
			$this->shop->update_option('wpsg_mod_abo_cron_run', 1);
			$this->shop->update_option('wpsg_mod_abo_cron_start', time());
			 
			@ini_set('memory_limit', '2000M');
			@set_time_limit(60 * 4);
			
            //$d1 = date_create('2022-04-20 12:05:00');
			//$d2 = date_create('2022-04-20 12:00:00'); // NOW
			//echo " = ".date_diff($d1, $d2)->format('%r%a');
			//die();
			
			$handledAboIDs = [];

			$arAboInformation = $this->db->fetchAssoc("
                SELECT
                    A.*
                FROM
                    `".WPSG_TBL_ABO."` AS A
                WHERE
                    A.`status` = '".wpsg_q(self::STATUS_ACTIVE)."' AND
                    A.`inform_autorenew` != '1'
            ");

            foreach ($arAboInformation as $a) {

            	try {

					$oOrderProduct = wpsg_order_product::getInstance(intval($a['order_product_id']));

				} catch (\Exception $e) {

					wpsg_debug($e->getMessage());
					continue;

				}

                $product_id = $oOrderProduct->getProductId();

                $wpsg_mod_abo_autoexpire_informationtime = intval($this->getSetting('wpsg_mod_abo_autoexpire_informationtime', $product_id)); // Benachrichtigung
                $interval = intval(date_diff(date_create($a['expiration']), date_create(date('Y-m-d H:i:s')))->format('%r%a')); // Tage bis zum auslaufen

	            // Positiv NOW() > Auslauf
 	            
                if ($wpsg_mod_abo_autoexpire_informationtime > 0 && abs($interval) < $wpsg_mod_abo_autoexpire_informationtime && $interval <= 0) {

                    $wpsg_mod_abo_expiremode = $this->getSetting('wpsg_mod_abo_premode', $product_id);

                    if ($wpsg_mod_abo_expiremode === self::EXPIRE_MAIL) {

                        $this->sendAboRenewInformation($a['order_id'], $a['order_product_id']);

                        $this->db->UpdateQuery(WPSG_TBL_ABO, array(
                            'inform_autorenew' => '1'
                        ), "`id` = '".wpsg_q($a['id'])."'");

                        $handledAboIDs[] = intval($a['id']);

                    } else if ($wpsg_mod_abo_expiremode === self::EXPIRE_AUTO) {
						
						//wpsg_debug("1");
                    	$this->autoExtendOrder($a['order_id'], $a['order_product_id'], intval($a['id']));
						$checkEndCron();

                    	$handledAboIDs[] = intval($a['id']);

                    }

                }

            }
 
			$arAboExpiration = $this->db->fetchAssoc("
				SELECT
					A.*
				FROM
					`".WPSG_TBL_ABO."` AS A
				WHERE
					A.`expiration` < NOW() AND
					A.`status` = '".wpsg_q(self::STATUS_ACTIVE)."' AND
					A.`inform_expiration` != '1' 					
			");

			foreach ($arAboExpiration as $a) {

				if (in_array(intval($a['id']), $handledAboIDs)) continue;

			    $oOrderProduct = wpsg_order_product::getInstance(intval($a['order_product_id']));
			    $product_id = $oOrderProduct->getProductId();

				// URL Benachrichtigung
				$url = $this->getNotifyURL($product_id);

				if (wpsg_isSizedString($url)) {

					$this->shop->notifyURL($url, $oOrderProduct->getProductKey(), 1, $a['order_id'], 3, array(
						'typ' => 'abo_expiration',
						'abo_kaufdatum' => date('Y-m-d H:i:s', time()),
						'abo_expiration' => $a['expiration']
					));

				}
                $wpsg_mod_abo_expiremode = $this->getSetting('wpsg_mod_abo_expiremode', $product_id);

                if ($wpsg_mod_abo_expiremode == self::EXPIRE_AUTO) {

					//wpsg_debug("2");
                    $this->autoExtendOrder($a['order_id'], $a['order_product_id'], intval($a['id']));
					$checkEndCron();
					
                } else {

				    // Mail mit Benachrichtigung senden
				    $this->sendAboExpirationMail($a['order_id'], $a['order_product_id']);

                }

				// Damit der Cron das nicht ständig macht
				$this->db->UpdateQuery(WPSG_TBL_ABO, array(
					'inform_expiration' => '1'
				), "`id` = '".wpsg_q($a['id'])."'");

			}

            $this->shop->update_option('wpsg_mod_abo_cron_lastrun', time());
			$this->shop->update_option('wpsg_mod_abo_cron_run', '0');

			unlink($pid_file);
			
			echo "Cron normal beendet<br />";
						
		}

		public function admin_emailconf_save()
		{

			wpsg_saveEMailConfig('wpsgmodabofirstbuy');
			wpsg_saveEMailConfig('wpsgmodaboexpiration');
			wpsg_saveEMailConfig('wpsgmodabrebuy');
            wpsg_saveEMailConfig('wpsgmodaboadminstoprenewal');
            wpsg_saveEMailConfig('wpsgmodabautoinfo');

		} // public function admin_emailconf_save()

		public function admin_emailconf() {

			echo wpsg_drawEMailConfig(
				'wpsgmodabofirstbuy',
				__('E-Mail bei erstmaligem Abo Kauf', 'wpsg'),
				__('Diese Mail bekommt der Kunde beim erstmaligen Kauf eines Abos.', 'wpsg'));

			echo wpsg_drawEMailConfig(
				'wpsgmodaboexpiration',
				__('E-Mail bei abgelaufenem Abo', 'wpsg'),
				__('Diese Mail bekommt der Kunde wenn ein Abo ausläuft mit der Möglichkeit es zu verlängern.', 'wpsg'));

			echo wpsg_drawEMailConfig(
				'wpsgmodabrebuy',
				__('E-Mail bei verlängertem Abo', 'wpsg'),
				__('Diese Mail bekommt der Kunde wenn er erfolgreich ein Abo verlängert hat.', 'wpsg'));

            echo wpsg_drawEMailConfig(
                'wpsgmodaboadminstoprenewal',
                __('E-Mail an Admin wenn Abo angehalten werden soll', 'wpsg'),
                __('Diese Mail wird verschickt, wenn ein Kunde im Frontend ein Abo anhalten will.', 'wpsg'), true);

            echo wpsg_drawEMailConfig(
            	'wpsgmodabautoinfo',
            	__('E-Mail bei Information vor Verlängerung', 'wpsg'),
            	__('Diese Mail erhalten Kunden, bevor ein Abo automatisch verlängert wird', 'wpsg'), true);

		} // public function admin_emailconf()

		public function template_redirect() {

			if (!isset($_REQUEST['mod']) || $_REQUEST['mod'] != 'wpsg_mod_abo') return false;

			if (!isset($_REQUEST['order_product_id']) || !isset($_REQUEST['order_id'])) {

				throw new \wpsg\Exception(__('Beim Verlängerungslink keine BestellID oder ProductKey übergeben', 'wpsg'));
				return false;

			}

			$product_key = 'abo_'.$_REQUEST['order_id'].'_'.$_REQUEST['order_product_id'];

			$oOrderProduct = wpsg_order_product::getInstance($_REQUEST['order_product_id']);

			// Kann das Abo verlängert werden?
			$product_data_db = $this->shop->cache->loadProduct($oOrderProduct->getProductId());

			// Wie oft wurde das Abo zu dieser Bestellung bereits verlängert?
			$nAboBuyed = $this->db->fetchOne("SELECT `rebuy` FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($_REQUEST['order_id'])."' AND `order_product_id` = '".wpsg_q($_REQUEST['order_product_id'])."'");

			// Ist so ein Abo Produkt schon im Warenkorb?
			$nDrin = false;
			
			if (wpsg_isSizedArray($_SESSION['wpsg']['basket'])) {

				foreach ($_SESSION['wpsg']['basket'] as $p_index => $p) {

					if ($p['id'] == $product_key) $nDrin = $p_index;

				}

			}

			if ($nDrin !== false) {

				if ($product_data_db['wpsg_mod_abo_resell'] != '0' && $nAboBuyed + $_SESSION['wpsg']['basket'][$nDrin]['menge'] + 1 >= $product_data_db['wpsg_mod_abo_resell']) {

					// Abo kann nicht mehr verlängert werden
					$this->shop->addFrontendError(__('Das Abo kann nicht mehr verlängert werden. Sie müssen das Produkt erneut erwerben.', 'wpsg'));
					$this->shop->redirect($this->shop->getURL(wpsg_ShopController::URL_BASKET));

				}

				$_SESSION['wpsg']['basket'][$nDrin]['menge'] ++;

			} else {

				if ($product_data_db['wpsg_mod_abo_resell'] != '0' && $nAboBuyed + 1 > $product_data_db['wpsg_mod_abo_resell']) {

					// Abo kann nicht mehr verlängert werden
					$this->shop->addFrontendError(__('Das Abo kann nicht mehr verlängert werden. Sie müssen das Produkt erneut erwerben.', 'wpsg'));
					$this->shop->redirect($this->shop->getURL(wpsg_ShopController::URL_BASKET));

				}

				if (!wpsg_isSizedArray($_SESSION['wpsg'])) $_SESSION['wpsg'] = array();
				if (!wpsg_isSizedArray($_SESSION['wpsg']['basket'])) $_SESSION['wpsg']['basket'] = array();

				$_SESSION['wpsg']['basket'][] = array(
					'menge' => '1',
					'id' => $product_key,
                    'product_key' => $product_key
				);

				$product_data = $this->shop->loadProduktArray($oOrderProduct->getProductId());

				$payment = $this->getSetting('wpsg_mod_abo_autoexpire_payment', $oOrderProduct->getProductId());

				if ($payment === 'origin') {

				    $_SESSION['wpsg']['checkout']['payment'] = $oOrderProduct->getOrder()->getPaymentID();

					/*
					if (intval($_SESSION['wpsg']['checkout']['payment']) === 20) {
						
						$_SESSION['wpsg']['checkout']['mod_autodebit_name'] = $oOrderProduct->getOrder()->__get('mod_autodebit_name');
						$_SESSION['wpsg']['checkout']['mod_autodebit_blz'] = $oOrderProduct->getOrder()->__get('mod_autodebit_blz');
						$_SESSION['wpsg']['checkout']['mod_autodebit_bic'] = $oOrderProduct->getOrder()->__get('mod_autodebit_bic');
						$_SESSION['wpsg']['checkout']['mod_autodebit_inhaber'] = $oOrderProduct->getOrder()->__get('mod_autodebit_inhaber');
						$_SESSION['wpsg']['checkout']['mod_autodebit_knr'] = $oOrderProduct->getOrder()->__get('mod_autodebit_knr');
						$_SESSION['wpsg']['checkout']['mod_autodebit_iban'] = $oOrderProduct->getOrder()->__get('mod_autodebit_iban');
												
					} 
					*/
					
                } else {

                    $_SESSION['wpsg']['checkout']['payment'] = $payment;

                }
				
				$shipping_adress_id = intval($oOrderProduct->getOrder()->__get('shipping_adress_id'));
				
				if ($shipping_adress_id > 0) {
				
					$row_adress = $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_ADRESS."` WHERE `id` = '".wpsg_q($shipping_adress_id)."' ");
					
					$_SESSION['wpsg']['checkout']['diff_shippingadress'] = '1';
					
					$_SESSION['wpsg']['checkout']['shipping_title'] = $row_adress['title'];
					$_SESSION['wpsg']['checkout']['shipping_plz'] = $row_adress['plz'];
					$_SESSION['wpsg']['checkout']['shipping_vname'] = $row_adress['vname'];
					$_SESSION['wpsg']['checkout']['shipping_name'] = $row_adress['name'];
					$_SESSION['wpsg']['checkout']['shipping_strasse'] = $row_adress['strasse'];
					$_SESSION['wpsg']['checkout']['shipping_nr'] = $row_adress['nr'];
					$_SESSION['wpsg']['checkout']['shipping_ort'] = $row_adress['ort'];
					$_SESSION['wpsg']['checkout']['shipping_land'] = $row_adress['land'];
					$_SESSION['wpsg']['checkout']['shipping_firma'] = $row_adress['firma'];
					$_SESSION['wpsg']['checkout']['shipping_firma'] = $row_adress['firma'];
						 	  					
				}

				$this->shop->addFrontendMessage(wpsg_translate(__('Abo Verlängerung für Produkt #1# eingefügt', 'wpsg'), $product_data['name']));

			}

			$this->shop->redirect($this->shop->getURL(wpsg_ShopController::URL_BASKET));

		} // public function template_redirect()

		public function basket_updateProduktFromSession(&$product_index, &$stock) {

            /*
            $product_id = $_SESSION['wpsg']['basket'][$product_index]['id'];

            $order_id = $this->getOrderIDFromKey($product_key);
            $product_id = $this->shop->getProduktID($product_key);
            $product_data = $this->shop->cache->loadProduct($product_id);

            $nAboBuyed = $this->db->fetchOne("SELECT `rebuy` FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."' AND `product_index` = '".wpsg_q($product_index)."'");

            if ($nAboBuyed + $stock > $product_data['wpsg_mod_abo_resell'] && $product_data['wpsg_mod_abo_resell'] != '0')
            {

                $this->shop->addFrontendError(__('Abo kann nicht mehr so oft verlängert werden!', 'wpsg'));
                $stock = 0;

            }
            */

		}

		public function basket_row(&$p, $i)
		{

			if (substr($p['id'], 0, 3) == 'abo')
			{

				$this->shop->view['wpsg_mod_abo']['order'] = $this->shop->cache->loadOrder($this->getOrderIDFromKey($p['id']));
				$this->shop->view['wpsg_mod_abo']['product'] = $this->shop->cache->loadProduct($this->shop->getProduktID($p['id']));
				$this->shop->view['wpsg_mod_abo']['wpsg_mod_abo_durration_menge'] = $p['menge'] * $this->shop->view['wpsg_mod_abo']['product']['wpsg_mod_abo_durration'];

				$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/basket_row.phtml');

			}

		} // public function basket_row(&$p, $i)

		public function overview_row(&$p, $i)
		{

			if (substr($p['id'], 0, 3) == 'abo')
			{

				$this->shop->view['wpsg_mod_abo']['order'] = $this->shop->cache->loadOrder($this->getOrderIDFromKey($p['id']));
				$this->shop->view['wpsg_mod_abo']['product'] = $this->shop->cache->loadProduct($this->shop->getProduktID($p['id']));
				$this->shop->view['wpsg_mod_abo']['wpsg_mod_abo_durration_menge'] = $p['menge'] * $this->shop->view['wpsg_mod_abo']['product']['wpsg_mod_abo_durration'];

				$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/overview_row.phtml');

			}

		} // public function overview_row(&$p, $i)

		public function mail_row($index, $produkt)
		{

			if (substr($produkt['id'], 0, 3) == 'abo')
			{

				$this->shop->view['wpsg_mod_abo']['order'] = $this->shop->cache->loadOrder($this->getOrderIDFromKey($produkt['id']));
				$this->shop->view['wpsg_mod_abo']['product'] = $this->shop->cache->loadProduct($this->shop->getProduktID($produkt['id']));
				$this->shop->view['wpsg_mod_abo']['wpsg_mod_abo_durration_menge'] = $produkt['menge'] * $this->shop->view['wpsg_mod_abo']['product']['wpsg_mod_abo_durration'];

				if ($this->shop->htmlMail === true)
				{

					return $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_row_html.phtml');

				}
				else
				{

					return $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_row.phtml');

				}

			}

		} // public function mail_row($index, $produkt)

		public function order_view_row(&$p, $i) {

			$this->shop->view['i'] = $i;

			if (substr(wpsg_getStr($p['product_key']), 0, 3) == 'abo') {

				$this->shop->view['wpsg_mod_abo']['order'] = $this->shop->cache->loadOrder($this->getOrderIDFromKey($p['product_key']));
				$this->shop->view['wpsg_mod_abo']['product'] = $this->shop->cache->loadProduct($this->shop->getProduktID($p['product_key']));
				$this->shop->view['wpsg_mod_abo']['wpsg_mod_abo_durration_menge'] = $p['amount'] * $this->shop->view['wpsg_mod_abo']['product']['wpsg_mod_abo_durration'];

				return $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/order_view_row.phtml');

			} else {

				$product_data = $this->shop->cache->loadProduct($p['product_id']);

				if ($product_data['wpsg_mod_abo_activ'] == '1')
				{

					$this->shop->view['wpsg_mod_abo']['rebuy'] = $this->db->fetchAssoc("SELECT * FROM `".WPSG_TBL_ORDERPRODUCT."` WHERE `productkey` LIKE 'abo_".wpsg_q($this->shop->view['data']['id']??'')."_%'");

					foreach ($this->shop->view['wpsg_mod_abo']['rebuy'] as $k => $v)
					{

						$product_data = $this->shop->cache->loadProduct($v['p_id']);

						$this->shop->view['wpsg_mod_abo']['rebuy'][$k]['product'] = $product_data;
						$this->shop->view['wpsg_mod_abo']['rebuy'][$k]['order'] = $this->shop->cache->loadOrder($v['o_id']);
						$this->shop->view['wpsg_mod_abo']['rebuy'][$k]['tage'] = $v['menge'] * $product_data['wpsg_mod_abo_durration'];

					}

					return $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/order_view_base.phtml');

				}

			}

		} // public function order_view_row(&$p, $i)

        public function be_ajax() {

            $action = $_REQUEST['subaction'].'Action';

            $this->$action();

        } // public function be_ajax()

        public function order_view($order_id, &$arSidebarArray)  {

            if ($this->isAboOrder($order_id) === 1) {

                $this->shop->view['arOrderedAboProducts'] = $this->getAboProductsFromOrder($order_id);
                $this->shop->view['wpsg_mod_abo'] = [
                    'oOrder' => wpsg_order::getInstance(intval($order_id))
                ];

                $arSidebarArray['abo'] = [
                    'title' => _('Abo Produkte'),
                    'content' => $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/order_view.phtml', false)
                ];

            }

        } // public function order_view($order_id, &$arSidebarArray)
 
        public function order_ajax() {

	        $_REQUEST['edit_id'] = intval($_REQUEST['edit_id']);
	        $_REQUEST['order_product_id'] = intval($_REQUEST['order_product_id']);

            if (wpsg_isSizedString($_REQUEST['do'], 'run')) {

                $oOrderProduct = wpsg_order_product::getInstance($_REQUEST['order_product_id']);

                $abo_status = $this->getAboProductStatus($_REQUEST['edit_id'], $_REQUEST['order_product_id']);
                $bNotify = $this->startRenewal($abo_status);
                $product_id = $oOrderProduct->getProductId();
                $wpsg_mod_abo_expiremode = $this->getSetting('wpsg_mod_abo_expiremode', $product_id);

                if ($wpsg_mod_abo_expiremode === self::EXPIRE_AUTO) $this->shop->addBackendMessage(__('Wenn das Abo ausläuft, wird wieder automatisch eine Bestellung ausgelöst.', 'wpsg'));
                else if ($wpsg_mod_abo_expiremode === self::EXPIRE_MAIL) $this->shop->addBackendMessage(__('Wenn das Abo ausläuft, wird dem Kunden wieder die Möglichkeit gegeben mit einer neuen Bestellung das Abo zu verlängern.', 'wpsg'));

                if ($bNotify) $this->shop->addBackendMessage(__('URL Benachrichtigung ausgelöst. (type=abo_status_started)', 'wpsg'));

                $this->shop->redirect(
                    wpsg_admin_url('Order', 'view', ['edit_id' => $_REQUEST['edit_id']])
                );

            } else if (wpsg_isSizedString($_REQUEST['do'], 'stop')) {

                $oOrderProduct = wpsg_order_product::getInstance($_REQUEST['order_product_id']);

                $abo_status = $this->getAboProductStatus($_REQUEST['edit_id'], $_REQUEST['order_product_id']);
                $bNotify = $this->stopRenewal($abo_status);
                $product_id = $oOrderProduct->getProductId();
                $wpsg_mod_abo_expiremode = $this->getSetting('wpsg_mod_abo_expiremode', $product_id);

                if ($wpsg_mod_abo_expiremode === self::EXPIRE_AUTO) $this->shop->addBackendMessage(__('Wenn das Abo ausläuft, wird keine Bestellung mehr ausgelöst.', 'wpsg'));
                else if ($wpsg_mod_abo_expiremode === self::EXPIRE_MAIL) $this->shop->addBackendMessage(__('Wenn das Abo ausläuft, wird dem Kunden keine Möglichkeit gegeben mit einer neuen Bestellung das Abo zu verlängern.', 'wpsg'));

                if ($bNotify) $this->shop->addBackendMessage(__('URL Benachrichtigung ausgelöst. (type=abo_status_stopped)', 'wpsg'));

                $this->shop->redirect(
                    wpsg_admin_url('Order', 'view', ['edit_id' => $_REQUEST['edit_id']])
                );

            }

        }

        public function wpsg_mod_export_loadFields(&$arFields) {

            $arFields[10]['fields']['wpsg_mod_abo_starttime'] = __('Abo Startzeit', 'wpsg');
            $arFields[10]['fields']['wpsg_mod_abo_runningtime'] = __('Abo Laufzeit', 'wpsg');
            $arFields[10]['fields']['wpsg_mod_abo_price'] = __('Abo Kosten Verlängerung', 'wpsg');
            $arFields[10]['fields']['wpsg_mod_abo_duration'] = __('Abo Dauer (Tage)', 'wpsg');

        }

        public function wpsg_mod_export_getValue(&$return, $field_value, $o_id, $p_id, $order_product_id, $profil_separator) {

            if (!in_array($field_value, ['wpsg_mod_abo_runningtime', 'wpsg_mod_abo_starttime', 'wpsg_mod_abo_price', 'wpsg_mod_abo_duration'])) return;

            $return = '';

            $abo_order = $this->isAboOrder($o_id);
            $abo_order_id = 0;
			
			$oOrderProduct = wpsg_order_product::getInstance($order_product_id);
			
			$product_key = $oOrderProduct->getProductKey();

            if ($abo_order === false) return;
            else if ($abo_order === 1) $abo_order_id = $o_id;
            else if ($abo_order === 2) $abo_order_id = $this->getOriginOrder($o_id);

            $arAboProducts = $this->getAboProductsFromOrder($abo_order_id);
            $arProductExportKeys = [];

            if (wpsg_isSizedInt($p_id)) {

                /** @var \wpsg_order_product $oOrderProduct */
                foreach ($arAboProducts as $oOrderProduct) {

                    if ($oOrderProduct->getProductKey() == $product_key) {

                        $arProductExportKeys[] = $product_key;

                    }

                }

            } else {

                /** @var \wpsg_order_product $oOrderProduct */
                foreach ($arAboProducts as $oOrderProduct) {

                    $arProductExportKeys[] = $oOrderProduct->getProductKey();

                }

            }

            if (wpsg_isSizedArray($arProductExportKeys)) {

                if ($field_value === 'wpsg_mod_abo_runningtime') {

                    foreach ($arProductExportKeys as $product_key) {

                        $abo_info = $this->getAboProductStatus($abo_order_id, $product_key);

                        $arReturn[] = wpsg_fromDate($abo_info['expiration']);

                    }

                    $return = implode(',', $arReturn);

                } else if ($field_value === 'wpsg_mod_abo_starttime') {

                    foreach ($arProductExportKeys as $product_key) {

                        $abo_info = $this->getAboProductStatus($abo_order_id, $product_key);

                        $arReturn[] = wpsg_fromDate($abo_info['cdate']);

                    }

                    $return = implode(',', $arReturn);

                } else if ($field_value === 'wpsg_mod_abo_price') {

                    foreach ($arProductExportKeys as $product_key) {

                        $oProduct = wpsg_product::getInstance($this->shop->getProduktID($product_key));

                        $arReturn[] = wpsg_ff($oProduct->wpsg_mod_abo_price);

                    }

                    $return = implode(',', $arReturn);

                } else if ($field_value === 'wpsg_mod_abo_duration') {

                    foreach ($arProductExportKeys as $product_key) {

                        $oProduct = wpsg_product::getInstance($this->shop->getProduktID($product_key));

                        $arReturn[] = $oProduct->wpsg_mod_abo_durration;

                    }

                    $return = implode(',', $arReturn);

                }

            }

        }

		/** Modulfunktionen */

        public function getSetting($key, $product_id = null) {

            if ($product_id !== null) {

                $oProduct = wpsg_product::getInstance($product_id);

                $meta_value = $oProduct->getMeta($key, false, self::CONFIG);

                if ($meta_value !== self::CONFIG) return $meta_value;

            }

            return $this->shop->get_option($key);

        }

		private function resetCronAction() {
			
			$this->shop->update_option('wpsg_mod_abo_cron_run', false);
			
			$upload_dir = wp_upload_dir();
			$pid_file = $upload_dir['basedir'].'/wpsg_cron_abo';
			
			unlink($pid_file);
			
			$this->shop->addBackendMessage(__('CRON Zurückgesetzt.', 'wpsg'));
			
			$this->shop->redirect(WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Admin&action=module&modul=wpsg_mod_abo');
			
		}
		
        private function admin_inlineEditAction() {

            $order_id = intval($_REQUEST['order_id']);
            $order_product_id = intval($_REQUEST['order_product_id']);
            $tDate = strtotime($_REQUEST['date']);

            if ($tDate <= 0) throw new \wpsg\Exception(__('Kein Datum übergeben.', 'wpsg'));

            $tDate += (3600 * get_option('gmt_offset')); // TODO: Klären warum das nicht richtig von x-editable geparsed wird, find es jetzt nicht

            $this->db->UpdateQuery(WPSG_TBL_ABO, [
                'expiration' => wpsg_q(date('Y-m-d H:i:s', $tDate))
            ], " `order_id` = '".wpsg_q($order_id)."' AND `order_product_id` = '".wpsg_q($order_product_id)."' ");

            $oAbo = wpsg_abo::getFromOrderAndOrderProduct($order_id, $order_product_id);

            wpsg_header::JSONData([
                'state' => $oAbo->getState(),
                'time' => wpsg_formatTimestamp($tDate)
            ]);

            exit;

        }

        /**
         * Verlängerung anhalten
         */
		private function stopRenewal($abo_status) {

            $this->db->UpdateQuery(WPSG_TBL_ABO, Array('status' => self::STATUS_STOPPED), " `id` = '".wpsg_q($abo_status['id'])."' ");

            $oOrderProduct = wpsg_order_product::getInstance(intval($abo_status['order_product_id']));

            $url = $this->getNotifyURL($oOrderProduct->getProductId());

            if (wpsg_isSizedString($url)) {

                $this->shop->notifyURL($url, $oOrderProduct->getProductKey(), 1, $abo_status['order_id'], 3, array(
                    'typ' => 'abo_status_stopp',
                    'abo_baseOrderID' => $abo_status['order_id']
                ));

                return true;

            }

            $oProduct = $oOrderProduct->getProduct();

            // Protokoll
            $oOrder = wpsg_order::getInstance($abo_status['order_id']);
            $oOrder->log(wpsg_translate(__('Abo für Produkt #1# wure angehalten.', 'wpsg'), $oProduct->getProductName()), '');

            return false;

        } // private function stopRenewal()

        /**
         * Verlängerung wieder aktivieren
         */
        private function startRenewal($abo_status)
        {

            $this->db->UpdateQuery(WPSG_TBL_ABO, Array(
                'status' => self::STATUS_ACTIVE,
                'inform_expiration' => '0'
            ), " `id` = '".wpsg_q($abo_status['id'])."' ");

            $oOrderProduct = wpsg_order_product::getInstance(intval($abo_status['order_product_id']));

            $url = $this->getNotifyURL($oOrderProduct->getProductId());

            if (wpsg_isSizedString($url))
            {

                $this->shop->notifyURL($url, $oOrderProduct->getProductKey(), 1, $abo_status['order_id'], 3, array(
                    'typ' => 'abo_status_started',
                    'abo_baseOrderID' => $abo_status['order_id']
                ));

                return true;

            }

            $oProduct = $oOrderProduct->getProduct();

            // Protokoll
            $oOrder = wpsg_order::getInstance($abo_status['order_id']);
            $oOrder->log(wpsg_translate(__('Abo für Produkt #1# wurde fortgesetzt.', 'wpsg'), $oProduct->getProductName()), '');

            return false;

        } // private function startRenewal($abo_status)

        /**
         * Frontend: Anhalten von Abo Verlängerungen
         */
        public function stoprenewalRedirect() {

	        $_REQUEST['order_id'] = wpsg_sinput("key", $_REQUEST['order_id']);
            $oOrder = wpsg_order::getInstance($_REQUEST['order_id']);

            // Sicherheitsüberprüfung
            if (!wpsg_isSizedInt($_SESSION['wpsg']['checkout']['id']) || $oOrder->k_id != $_SESSION['wpsg']['checkout']['id']) throw new \wpsg\Exception(__('Nicht angemeldet oder kein Zugriff auf diese Bestellung.', 'wpsg'));

            $arOrderProducts = $this->getAboProductsFromOrder($oOrder->id);

            foreach ($arOrderProducts as $oOrderProduct) {

                $abo_status = $this->getAboProductStatus($oOrder->id, $oOrderProduct->getId());

                if ($oOrderProduct->getProductKey() === $_REQUEST['product_key'] && wpsg_isSizedInt($abo_status['id'])) {

                    $wpsg_mod_abo_autoexpire_customer = $this->getSetting('wpsg_mod_abo_autoexpire_customer', $oOrderProduct->getProductId());

                    // Hier ist alles ok, das Abo kann gestoppt werden
                    if ($wpsg_mod_abo_autoexpire_customer === wpsg_mod_abo::CUSTOMER_AUTO_DIRECT) {

                        $this->stopRenewal($abo_status);

                        $this->shop->addFrontendMessage(__('Die Verlängerung ihres Abos wurde angehalten.', 'wpsg'));
                        $this->shop->redirect($this->shop->getURL(wpsg_ShopController::URL_ABO));

                    }
                    else if ($wpsg_mod_abo_autoexpire_customer === wpsg_mod_abo::CUSTOMER_AUTO_MAIL)
                    {

                        try
                        {

                            $this->sendAdminMailStopRenewal($abo_status);

                            $this->shop->addFrontendMessage(__('Der Administrator wurde informiert und wird die Verlängerung ihres Abos veranlassen.', 'wpsg'));

                        }
                        catch (\wpsg\Exception $ex)
                        {

                            $this->shop->addFrontendMessage($ex->getMessage());

                        }

                        $this->shop->redirect($this->shop->getURL(wpsg_ShopController::URL_ABO));

                    }
                    else throw new \wpsg\Exception(__('Funktion ist deaktiviert.', 'wpsg'));

                }

            }

            throw new \wpsg\Exception(__('Produkt nicht in Bestellung gefunden.', 'wpsg'));

        } // public function stoprenewalRedirect()

        /**
         * Sendet eine Mail an den Admin, wenn ein Kunde ein Abo anhalten möchte
         *
         * @param $abo_status
         */
        private function sendAdminMailStopRenewal($abo_status)
        {

            if (strtotime($abo_status['mail_admin_stop']) > 0) throw new \wpsg\Exception(wpsg_translate(__('Admin wurde bereits am #1# benachrichtigt das Abo zu stoppen.'), wpsg_formatTimestamp($abo_status['mail_admin_stop'], false)));

            $this->db->UpdateQuery(WPSG_TBL_ABO, Array(
                'mail_admin_stop' => 'NOW()'
            ), " `id` = '".wpsg_q($abo_status['id'])."' ");

            // Mail senden
            $this->shop->view['oOrder'] = wpsg_order::getInstance($abo_status['order_id']);
            $this->shop->view['oCustomer'] = $this->shop->view['oOrder']->getCustomer();
            $this->shop->view['oProduct'] = wpsg_product::getInstance($abo_status['product_key']);
            $this->shop->view['product_key'] = $abo_status['product_key'];

            if ($this->shop->get_option('wpsg_htmlmail') === '1') $mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_admin_stoprenewal_html.phtml', false);
            else $mail_html = false;

            $mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_admin_stoprenewal.phtml', false);

            $to = $this->shop->get_option('wpsg_wpsgmodaboadminstoprenewal_empfaenger');

            $oProduct = wpsg_product::getInstance($this->shop->getProduktID($abo_status['product_key']));

            // Protokoll
            $oOrder = wpsg_order::getInstance($abo_status['order_id']);
            $oOrder->log(wpsg_translate(__('Mail an Administrator (Abo anhalten)', 'wpsg'), $oProduct->getProductName()), $mail_text);

            $this->shop->sendMail($mail_text, $to, 'wpsgmodaboadminstoprenewal', array(), $abo_status['order_id'], false, $mail_html);

        } // private function sendAdminMailStopRenewal($abo_status)

        /**
         * Sendet eine Mail an den Kunden, wenn der Admin ein Abo gekündigt hat
         *
         * @param $abo_status
         */
        private function sendCustomerMailResign($abo_status)
        {

        	if (strtotime($abo_status['mail_admin_stop']) > 0) throw new \wpsg\Exception(wpsg_translate(__('Admin wurde bereits am #1# benachrichtigt das Abo zu stoppen.'), wpsg_formatTimestamp($abo_status['mail_admin_stop'], false)));

        	$this->db->UpdateQuery(WPSG_TBL_ABO, Array(
        			'mail_admin_stop' => 'NOW()'
        			), " `id` = '".wpsg_q($abo_status['id'])."' ");

        	// Mail senden
        	$this->shop->view['oOrder'] = wpsg_order::getInstance($abo_status['order_id']);
        	$this->shop->view['oCustomer'] = $this->shop->view['oOrder']->getCustomer();
        	$this->shop->view['oProduct'] = wpsg_product::getInstance($abo_status['product_key']);
        	$this->shop->view['product_key'] = $abo_status['product_key'];

        	if ($this->shop->get_option('wpsg_htmlmail') === '1') $mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_storno_html.phtml', false);
        	else $mail_html = false;

        	$mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_storno.phtml', false);

        	//$to = $this->shop->get_option('wpsg_wpsgmodaboadminstoprenewal_empfaenger');
        	$to = $this->shop->view['customer']['email'];

        	$oProduct = wpsg_product::getInstance($this->shop->getProduktID($abo_status['product_key']));

        	// Protokoll
        	$oOrder = wpsg_order::getInstance($abo_status['order_id']);
        	$oOrder->log(wpsg_translate(__('Mail an Kunden (Abo gekündigt)', 'wpsg'), $oProduct->getProductName()), $mail_text);

        	$this->shop->sendMail($mail_text, $to, 'wpsgmodaboadminresign', array(), $abo_status['order_id'], false, $mail_html);

        } // private function sendCustomerMailResign($abo_status)

        /**
         * Erstellt automatisch eine Bestellung wenn ein Abo ausgelaufen ist
         *
         * @param $order_id
         * @param $product_key
         */
		public function autoExtendOrder($order_id, $order_product_id, $abo_id) {
		
		    $oOrderProduct = wpsg_order_product::getInstance($order_product_id);
            $product_id = $oOrderProduct->getProductId();

			$strQuery = "
				SELECT
					COUNT(*)
				FROM
					`".WPSG_TBL_ORDERPRODUCT."` AS OP
						LEFT JOIN `".WPSG_TBL_ORDER."` AS O ON (O.`id` = OP.`o_id`)
				WHERE
					OP.`p_id` = '".wpsg_q($product_id)."' AND
					OP.`productkey` = '".wpsg_q('abo_'.$order_id.'_'.$order_product_id)."' AND
					O.`payed_date` = '0000-00-00 00:00:00' AND
					O.`status` != '".wpsg_ShopController::STATUS_STORNIERT."' 					
			";
			
            $order_extend_exist = intval($this->db->fetchOne($strQuery));
			
			//wpsg_debug('....');
			//wpsg_debug("Verlängere1: ".$order_id.":".$order_product_id);
			
            if ($order_extend_exist > 0) return;

			 $this->shop->view = [];
			
			if (isset($_REQUEST['allinkl'])) echo "Verlängere Bestellung ".$order_id."<br />"; 
			else print_r($_REQUEST);
			
			$GLOBALS['wpsg_mod_abo_cron_count'] = ($GLOBALS['wpsg_mod_abo_cron_count']??0) + 1; 
			
			//wpsg_debug("Verlängere1.5: ".$order_id.":".$order_product_id);
			
		    if (!is_numeric($product_id)) throw new \Exception(__('Nicht vorgesehen!', 'wpsg'));

            $this->shop->addShipPay();

			//wpsg_debug("Verlängere2: ".$order_id.":".$order_product_id);
			
            $oOrder = wpsg_order::getInstance($order_id);

			//wpsg_debug("Verlängere3: ".$order_id.":".$order_product_id.":".$oOrder->getId());
		
            unset($_SESSION['wpsg']['basket']);
            unset($_SESSION['wpsg']['checkout']);

            $_SESSION['wpsg']['checkout']['id'] = $oOrder->k_id;

            $payment = $this->getSetting('wpsg_mod_abo_autoexpire_payment', $product_id);

            if ($payment === 'origin') {

                $_SESSION['wpsg']['checkout']['payment'] = $oOrderProduct->getOrder()->getPaymentID();

            } else {

                $_SESSION['wpsg']['checkout']['payment'] = $payment;

            }

            $_SESSION['wpsg']['checkout']['shipping'] = $oOrder->type_shipping;

            $customer_db_data = $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_KU."` WHERE `id` = '".wpsg_q($oOrder->k_id)."' ");
            $customer_address = $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_ADRESS."` WHERE `id` = '".wpsg_q($customer_db_data['adress_id'])."' ");

            $_SESSION['wpsg']['checkout']['firma'] = wpsg_getStr($customer_address['firma']);
            $_SESSION['wpsg']['checkout']['title'] = wpsg_getStr($customer_address['title']);
            $_SESSION['wpsg']['checkout']['vname'] = wpsg_getStr($customer_address['vname']);
            $_SESSION['wpsg']['checkout']['name'] = wpsg_getStr($customer_address['name']);
            $_SESSION['wpsg']['checkout']['email'] = wpsg_getStr($customer_db_data['email']);
            $_SESSION['wpsg']['checkout']['geb'] = wpsg_getStr($customer_address['geb']);
            $_SESSION['wpsg']['checkout']['fax'] = wpsg_getStr($customer_address['fax']);
            $_SESSION['wpsg']['checkout']['tel'] = wpsg_getStr($customer_address['tel']);
            $_SESSION['wpsg']['checkout']['strasse'] = wpsg_getStr($customer_address['strasse']);
            $_SESSION['wpsg']['checkout']['nr'] = wpsg_getStr($customer_address['nr']);
            $_SESSION['wpsg']['checkout']['plz'] = wpsg_getStr($customer_address['plz']);
            $_SESSION['wpsg']['checkout']['ort'] = wpsg_getStr($customer_address['ort']);
            $_SESSION['wpsg']['checkout']['land'] = wpsg_getStr($customer_address['land']);

			\wpsg\wpsg_calculation::$functionscache = [];
			
            $oBasket = new wpsg_basket();

            $oBasket->addProduktToSession('abo_'.$order_id.'_'.$order_product_id, 1);
			
			//wpsg_debug("Verlängere4: ".$order_id.":".$order_product_id.":".$oOrder->getId()."=".'abo_'.$order_id.'_'.$order_product_id);
			
            $oBasket->initFromSession(true); $arBasket = $oBasket->toArray();
            $new_o_id = $oBasket->save(true, false);

			$db_update = [
				'type_shipping' => wpsg_q($oOrder->type_shipping),
				'shipping_key' => wpsg_q($oOrder->shipping_key),
				'shipping_set' => wpsg_q($oOrder->shipping_set)
			];
			
			$order_data_old = $this->shop->cache->loadOrder($order_id);
			
			if ($this->shop->hasMod('wpsg_mod_autodebit')) {
						 
				$arBasket['checkout']['mod_autodebit_name'] = wpsg_q($order_data_old['mod_autodebit_name']);
				$arBasket['checkout']['mod_autodebit_blz'] = wpsg_q($order_data_old['mod_autodebit_blz']);
				$arBasket['checkout']['mod_autodebit_bic'] = wpsg_q($order_data_old['mod_autodebit_bic']);
				$arBasket['checkout']['mod_autodebit_inhaber'] = wpsg_q($order_data_old['mod_autodebit_inhaber']);
				$arBasket['checkout']['mod_autodebit_knr'] = wpsg_q($order_data_old['mod_autodebit_knr']);
				$arBasket['checkout']['mod_autodebit_iban'] = wpsg_q($order_data_old['mod_autodebit_iban']);
				
				$db_update['mod_autodebit_name'] = wpsg_q($order_data_old['mod_autodebit_name']);
				$db_update['mod_autodebit_blz'] = wpsg_q($order_data_old['mod_autodebit_blz']);
				$db_update['mod_autodebit_bic'] = wpsg_q($order_data_old['mod_autodebit_bic']);
				$db_update['mod_autodebit_inhaber'] = wpsg_q($order_data_old['mod_autodebit_inhaber']);
				$db_update['mod_autodebit_knr'] = wpsg_q($order_data_old['mod_autodebit_knr']);
				$db_update['mod_autodebit_iban'] = wpsg_q($order_data_old['mod_autodebit_iban']);
								
			}
			
			if ($order_data_old['adress_id'] !== $order_data_old['shipping_adress_id']) {
				
				$db_update['shipping_adress_id'] = $order_data_old['shipping_adress_id'];
				
				$arBasket['checkout']['shipping_firma'] = $oOrder->getShippingCompany();
				$arBasket['checkout']['shipping_vname'] = $oOrder->getShippingFirstName();
				$arBasket['checkout']['shipping_name'] = $oOrder->getShippingName();
				$arBasket['checkout']['shipping_strasse'] = $oOrder->getShippingStreet();
				$arBasket['checkout']['shipping_nr'] = $oOrder->getShippingStreetNr();
				$arBasket['checkout']['shipping_plz'] = $oOrder->getShippingZip();
				$arBasket['checkout']['shipping_ort'] = $oOrder->getShippingCity();
				$arBasket['shipping_land'] = [
					'name' => $oOrder->getShippingCountryName()
				];
				
			}
			
			$this->db->UpdateQuery(WPSG_TBL_ORDER, $db_update, " `id` = '".wpsg_q($new_o_id)."' ");
			
			wpsg_order::clearCache($new_o_id);
			$this->shop->cache->clearOrderCache($new_o_id);
			
            $arBasket['checkout']['knr'] = $customer_db_data['knr'];
            $arBasket['checkout']['onr'] = $this->db->fetchOne("SELECT `onr` FROM `".WPSG_TBL_ORDER."` WHERE `id` = '".wpsg_q($new_o_id)."' ");
            $arBasket['checkout']['datum'] = time();

            $bCustomerMail = false;
            $bAdminMail = false;

            if ($this->getSetting('wpsg_mod_abo_autoexpire_customermail', $product_id) === '1') $bCustomerMail = true;
            if ($this->getSetting('wpsg_mod_abo_autoexpire_adminmail', $product_id) === '1') $bAdminMail = true;

			//wpsg_debug("Verlängere5: ".$order_id.":".$order_product_id.":".$oOrder->getId());
			
            $oOrder->addLogEntry(__('Verlängerung erworben', 'wpsg'), 'OrderID: '.$order_id.'/'.$order_product_id.'/'.$product_id.'/'.$new_o_id.'/'.$abo_id.'/'.getmypid());

            $oBasket->sendOrderSaveMails($new_o_id, $arBasket, $bCustomerMail, $bAdminMail);

        }

        /**
         * Gibt einen Array mit AboProdukten einer Bestellung zurück
         * @param int $order_id
         * @return wpsg_order_product[]
         */
        public function getAboProductsFromOrder($order_id) {

            $oOrder = wpsg_order::getInstance($order_id);
            $arOrderProducts = Array();

            foreach ($oOrder->getOrderProducts() as $oOrderProduct) {

                if ($oOrderProduct->getProduct()->wpsg_mod_abo_activ) {

                    $arOrderProducts[] = $oOrderProduct;

                }

            }

            return $arOrderProducts;

        } // public function getAboProductsFromOrder($order_id)

        /**
         * Gibt die Informationen zu dem Abo aus der Abo Tabelle zurück
         *
         * @param $order_id
         * @param $product_key
         * @return mixed
         */
        public function getAboProductStatus($order_id, $order_product_id) {

            $abo_product = $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."' AND `order_product_id` = '".wpsg_q($order_product_id)."' ");

            return $abo_product;

        }

		/**
		 * Gibt das Datum im MySQL Format zurück, wann das Abo abläuft wenn es heute gekauft wird
		 * @param int $product_id Produkt ID
		 * @param int $menge Anzahl an gekauften Produkten
		 */
		public function getAboExpiration($product_id, $menge, $start, $noTimeZoneOffset = false) {

			$product_data = $this->shop->cache->loadProduct($product_id);

			if ($noTimeZoneOffset) return date('Y-m-d H:i:s', $start + $product_data['wpsg_mod_abo_durration'] * 24 * 60 * 60 * $menge);
			else return wpsg_date('Y-m-d H:i:s', $start + $product_data['wpsg_mod_abo_durration'] * 24 * 60 * 60 * $menge);

		} // public function getAboExpiration($product_id, $menge)

        public function isAboExtensionProductKey($product_key) {

		    if (preg_match('/^abo\_\d+\_\d+$/', $product_key)) {

		        return true;

            } else return false;

        }

        /**
         * Gibt die ProduktID aus einem Abo Verlängerungsschlüssel zurück
         * abo_{order_id}_{order_product_id}
         * @param $productkey
         */
        public function getProductIDFromAboExtensionProductKey($product_key) {

		    $arKey = explode('_', $product_key);

		    $oOrderProduct = wpsg_order_product::getInstance(intval($arKey[2]));

		    return $oOrderProduct->getProductId();

        }

		/**
		 * Gibt false zurück wenn die Bestellung keine Abo Bestellung ist
		 * 1 - Originalbestellung
		 * 2 - Folgebestellung
		 */
		public function isAboOrder($order_id) {

            // Folgebestellung testen
            $nFolgeBestellung = $this->db->fetchOne("SELECT COUNT(*) FROM `".WPSG_TBL_ORDERPRODUCT."` WHERE `o_id` = '".wpsg_q($order_id)."' AND `productkey` REGEXP '^abo_[0-9]+_[0-9]+' ");
            if ($nFolgeBestellung > 0) return 2;

			// Bei Originalbestellungen gibt es einen eintrag in der Tabelle
			$nAboBestellung = $this->db->fetchOne("SELECT COUNT(*) FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."'");
			if ($nAboBestellung > 0) return 1;

			return false;

		} // public function isAboOrder($order_id)

		/**
		 * Gibt die BestellID der Originalbestellung einer Verlängerung zurück
		 */
		public function getOriginOrder($order_id) {

			if ($this->isAboOrder($order_id) !== 2) return false;

			$productkeys = $this->db->fetchAssocField("SELECT `productkey` FROM `".WPSG_TBL_ORDERPRODUCT."` WHERE `o_id` = '".wpsg_q($order_id)."' ");

			foreach ($productkeys as $pk) {

				if (preg_match('/^abo_\d+_\d+$/', $pk)) {

					$arAbo = explode('_', $pk);

					return $arAbo[1];

				}

			}

			return false;

		} // public function getOriginOrder($order_id)

		/**
		 * Gibt das Auslaufdatum einer Abobestellung zurück
         *
		 * @param unknown_type $order_id
		 */
		public function getOrderAboExpiration($order_id, $order_product_id) {

			$data = null;

			$abo = wpsg_abo::getFromOrderAndOrderProduct($order_id, $order_product_id);

			$data = $abo->__get('expiration');

			/*
			$data = wpsg_fromDate($this->db->fetchOne("
                SELECT
                    `expiration`
                FROM
                    `".WPSG_TBL_ABO."`
                WHERE
                    `order_id` = '".wpsg_q($order_id)."' AND
                    `order_product_id` = '".wpsg_q($order_product_id)."'
            "), false);
			*/

			return $data;

		} // public function getOrderAboExpiration($order_id)

		/**
		 * Gibt die Benachrichtigungsurl für ein Produkt zurück
		 * @param int $product_id ProduktID
		 */
		public function getNotifyURL($product_id)
		{

			$product_data = $this->shop->cache->loadProduct($product_id);

			if (wpsg_isSizedString($product_data['wpsg_mod_abo_url'])) return $product_data['wpsg_mod_abo_url'];
			else return $this->shop->get_option('wpsg_mod_abo_url');

		} // public function getNotifyURL($product_id)

		/**
		 * Sendet die Mail beim erstmaligen Kauf eines Abo Produktes
		 * @param int $order_key BestellID
		 * @param String $product_id ProduktKey
		 */
		public function sendAboFirstBuyMail($order_id, $order_product_id) {

			if ($this->shop->get_option('wpsg_mod_abo_wpsgmodabofirstbuy') != '1') return;
            
			$oOrderProduct = wpsg_order_product::getInstance($order_product_id);
            
			$product_data = $this->shop->cache->loadProduct($oOrderProduct->getProductId());
			$order_data = $this->shop->cache->loadOrder($order_id);
			$customer_data = $this->shop->cache->loadKunden($order_data['k_id']);

			$this->shop->view['product'] = $product_data;
			$this->shop->view['customer'] = $customer_data;
			$this->shop->view['order'] = $order_data;
			$this->shop->view['rebuyLink'] = $this->getRebuyLink($order_id, $order_product_id);

			if ($this->shop->get_option('wpsg_htmlmail') === '1')
			{

				$mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_firstbuy_html.phtml', false);

			}
			else
			{

				$mail_html = false;

			}

			$mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_firstbuy.phtml', false);

			$this->shop->sendMail($mail_text, $customer_data['email'], 'wpsgmodabofirstbuy', array(), $order_id, $customer_data['id'], $mail_html);

			return true;

		} // public function sendAboFirstBuyMail($order_id, $product_id)

        /**
         * Sendet die Mail, die X-Tage vor der automatischen Verlängerung versendet werden kann
         * @param int $order_id BestellID
         * @param String $product_key ProduktKey
         */
        public function sendAboRenewInformation($order_id, $order_product_id) {

            $oOrderProduct = wpsg_order_product::getInstance($order_product_id);
            
            $product_id = $oOrderProduct->getProductId();

            if ($this->getSetting('wpsg_mod_abo_premode', $product_id) !== self::EXPIRE_MAIL) return;
            
            $product_data = $this->shop->cache->loadProduct($product_id);
            $order_data = $this->shop->cache->loadOrder($order_id);
            $customer_data = $this->shop->cache->loadKunden($order_data['k_id']);
                        
            $this->shop->view['oProduct'] = wpsg_product::getInstance($product_data['id']);
            $this->shop->view['product'] = $product_data;
            $this->shop->view['customer'] = $customer_data;
            $this->shop->view['order'] = $order_data;
            $this->shop->view['wpsg_mod_abo_autoexpire_informationtime'] = $this->getSetting('wpsg_mod_abo_autoexpire_informationtime', $product_id);
            $this->shop->view['rebuyLink'] = $this->getRebuyLink($order_id, $order_product_id);

            if ($this->shop->get_option('wpsg_htmlmail') === '1') $mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_autoinformation_html.phtml', false);
            else $mail_html = false;

            $mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_autoinformation.phtml', false);
 
            $this->shop->sendMail($mail_text, $customer_data['email'], 'wpsgmodabautoinfo', array(), $order_id, $customer_data['id'], $mail_html);
            
            $oOrder = wpsg_order::getInstance($order_id);
            $oOrder->addLogEntry(__('Benachrichtigung Abo Ablauf', 'wpsg'), $mail_text);

            return true;
            
        } // public function sendAboRenewInformation($order_id, $product_key)
        
		/**
		 * Sendet die Mail bei ablauf eines Abos
		 * @param int $order_key BestellID
		 * @param String $product_id ProduktKey
		 * @return boolean
		 */
		public function sendAboExpirationMail($order_id, $order_product_id) {

		    $oOrderProduct = wpsg_order_product::getInstance($order_product_id);
		    $oOrder = wpsg_order::getInstance($order_id);
		    
		    $product_id = $oOrderProduct->getProductId();
		    $wpsg_mod_abo_expiremode = $this->getSetting('wpsg_mod_abo_expiremode', $product_id);
		    
			if ($wpsg_mod_abo_expiremode != self::EXPIRE_MAIL) return;

			$this->shop->view = [];
			
			$product_data = $this->shop->cache->loadProduct($product_id);
			$order_data = $this->shop->cache->loadOrder($order_id);
			$customer_data = $this->shop->cache->loadKunden($order_data['k_id']);

			$this->shop->view['product'] = $product_data;
			$this->shop->view['customer'] = $customer_data;
			$this->shop->view['order'] = $order_data;
			$this->shop->view['rebuyLink'] = $this->getRebuyLink($order_id, $order_product_id);
			$this->shop->view['allowRebuy'] = $this->canReBuyAbo($order_id, $order_product_id);

			if ($this->shop->get_option('wpsg_htmlmail') === '1') {

				$mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_expiration_html.phtml', false);

			} else {

				$mail_html = false;

			}

			$mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_expiration.phtml', false);

			$this->shop->sendMail($mail_text, $customer_data['email'], 'wpsgmodaboexpiration', array(), $order_id, $customer_data['id'], $mail_html);
            $oOrder->addLogEntry(__('Abo Ablauf', 'wpsg'), $mail_text);

			return true;

		} // public function sendAboExpirationMail($order_id, $product_id)

		/**
		 * Sendet die Mail bei Verlängerung eines Abos
		 * @param int $order_key BestellID
		 * @param String $product_id ProduktKey
		 * @return boolean
		 */
		public function sendAboRebuyMail($order_id, $order_product_id) {

			if ($this->shop->get_option('wpsg_mod_abo_wpsgmodabrebuy') != '1') return;

			$oOrderProduct = wpsg_order_product::getInstance($order_product_id);
						
			$product_data = $this->shop->cache->loadProduct($oOrderProduct->getProductId());
			$order_data = $this->shop->cache->loadOrder($order_id);
			$customer_data = $this->shop->cache->loadKunden($order_data['k_id']);

			$this->shop->view['product'] = $product_data;
			$this->shop->view['customer'] = $customer_data;
			$this->shop->view['order'] = $order_data;
			$this->shop->view['rebuyLink'] = $this->getRebuyLink($order_id, $order_product_id);
			$this->shop->view['allowRebuy'] = $this->canReBuyAbo($order_id, $order_product_id);

			if ($this->shop->get_option('wpsg_htmlmail') === '1') {

				$mail_html = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_rebuy_html.phtml', false);

			} else {

				$mail_html = false;

			}

			$mail_text = $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_abo/mail_rebuy.phtml', false);

			$this->shop->sendMail($mail_text, $customer_data['email'], 'wpsgmodabrebuy', array(), $order_id, $customer_data['id'], $mail_html);

			return true;

		} // public function sendAboRebuyMail($order_id, $product_key)
 
		/**
		 * Gibt false zurück wenn das Abo nicht mehr verlängert werden kann, ansonsten true oder die Anzahl wenn Anzahl beschränkt die noch möglich ist
		 * @param int $order_id BestellID
		 * @param String $product_key ProduktKey
		 */
		public function canReBuyAbo($order_id, $order_product_id) {
		    
		    $oOrderProduct = wpsg_order_product::getInstance($order_product_id);

			$product_data = $this->shop->cache->loadProduct($oOrderProduct->getProductKey());

			if ($product_data['wpsg_mod_abo_resell'] == '0') {

				return true;

			} else {

				$rebuy = $this->db->fetchOne("SELECT `rebuy` FROM `".WPSG_TBL_ABO."` WHERE `order_id` = '".wpsg_q($order_id)."' AND `order_product_id` = '".wpsg_q($order_product_id)."'");

				if ($rebuy < $product_data['wpsg_mod_abo_resell'])
				{

					return $product_data['wpsg_mod_abo_resell'] - $rebuy;

				}

			}

			return false;

		} // public function canReBuyAbo($order_id, $product_key)

        /**
         * Gibt aus einem Produktschlüssel die BestellID zurück
         * Hier ist der Produktschlüssel z.b. abo_{original_order_id}_{original_order_product_id}
         * @param varchar ProduktKey
         */
        public function getOrderProductIDFromKey($product_key) {

            $arData = explode('_', $product_key);

            return wpsg_getInt($arData[2]);
            
        }
        
		/**
		 * Gibt aus einem Produktschlüssel die BestellID zurück
         * Hier ist der Produktschlüssel z.b. abo_{original_order_id}_{original_order_product_id}
		 * @param varchar ProduktKey
		 */
		public function getOrderIDFromKey($product_key) {

			$arData = explode('_', $product_key);

			return wpsg_getInt($arData[1]);

		} // public function getOrderIDFromKey($product_key)

		/**
		 * Gibt den Link zum wiederkauf eines Abos zurück
		 * @param int $order_id BestellID
		 * @param int $product_key ProduktKey
		 */
		public function getRebuyLink($order_id, $order_product_id) {
		    
		    $oOrderProduct = wpsg_order_product::getInstance($order_product_id);

			$url = get_permalink($this->shop->get_option('wpsg_page_basket'));

			$url .= ((strpos($url, '?') === false)?'?mod=wpsg_mod_abo':'&mod=wpsg_mod_abo');
			$url .= ((strpos($url, '?') === false)?'?order_id=':'&order_id=').$order_id;
			$url .= ((strpos($url, '?') === false)?'?product_key=':'&order_product_id=').$order_product_id;

			return $url;

		} // public function getRebuyLink($order_id, $product_key)

	} // class wpsg_mod_abo extends wpsg_mod_basic
 